﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

//#define EXPORT_GRAPH
//#define LOG_CONSOLE
//#define LOG_FILE

using System;
using System.Collections.Generic;
#if LOG_CONSOLE
using System.Diagnostics;
#endif
#if LOG_FILE
using System.Globalization;
using System.IO;
#endif
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HEAL.Attic;
using Tensorflow;
using Tensorflow.NumPy;
using static Tensorflow.Binding;
using static Tensorflow.KerasApi;
using DoubleVector = MathNet.Numerics.LinearAlgebra.Vector<double>;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [StorableType("63944BF6-62E5-4BE4-974C-D30AD8770F99")]
  [Item("TensorFlowConstantOptimizationEvaluator", "")]
  public class TensorFlowConstantOptimizationEvaluator : SymbolicRegressionConstantOptimizationEvaluator {
    private const string MaximumIterationsName = "MaximumIterations";
    private const string LearningRateName = "LearningRate";

    //private static readonly TF_DataType DataType = tf.float64;
    private static readonly TF_DataType DataType = tf.float32;

    #region Parameter Properties
    public IFixedValueParameter<IntValue> ConstantOptimizationIterationsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[MaximumIterationsName]; }
    }
    public IFixedValueParameter<DoubleValue> LearningRateParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters[LearningRateName]; }
    }
    #endregion

    #region Properties
    public int ConstantOptimizationIterations {
      get { return ConstantOptimizationIterationsParameter.Value.Value; }
    }
    public double LearningRate {
      get { return LearningRateParameter.Value.Value; }
    }
    #endregion

    public TensorFlowConstantOptimizationEvaluator()
      : base() {
      Parameters.Add(new FixedValueParameter<IntValue>(MaximumIterationsName, "Determines how many iterations should be calculated while optimizing the constant of a symbolic expression tree(0 indicates other or default stopping criterion).", new IntValue(10)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(LearningRateName, "", new DoubleValue(0.001)));
    }

    protected TensorFlowConstantOptimizationEvaluator(TensorFlowConstantOptimizationEvaluator original, Cloner cloner)
      : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TensorFlowConstantOptimizationEvaluator(this, cloner);
    }

    [StorableConstructor]
    protected TensorFlowConstantOptimizationEvaluator(StorableConstructorFlag _) : base(_) { }

    protected override ISymbolicExpressionTree OptimizeConstants(
      ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows,
      CancellationToken cancellationToken = default(CancellationToken), EvaluationsCounter counter = null) {
      return OptimizeTree(tree,
        problemData, rows,
        ApplyLinearScalingParameter.ActualValue.Value, UpdateVariableWeights,
        ConstantOptimizationIterations, LearningRate,
        cancellationToken);
    }

    public static ISymbolicExpressionTree OptimizeTree(ISymbolicExpressionTree tree,
      IRegressionProblemData problemData, IEnumerable<int> rows,
      bool applyLinearScaling, bool updateVariableWeights, int maxIterations, double learningRate,
      CancellationToken cancellationToken = default(CancellationToken), IProgress<double> progress = null) {

      const bool eager = true;

     bool prepared = TreeToTensorConverter.TryPrepareTree(
        tree,
        problemData, rows.ToList(),
        updateVariableWeights, applyLinearScaling,
        eager,
        out Dictionary<string, Tensor> inputFeatures, out Tensor target,
        out Dictionary<ISymbolicExpressionTreeNode, ResourceVariable[]> variables);
      if (!prepared)
        return (ISymbolicExpressionTree)tree.Clone();

      var optimizer = keras.optimizers.Adam((float)learningRate);
      
      for (int i = 0; i < maxIterations; i++) {
        if (cancellationToken.IsCancellationRequested) break;

        using var tape = tf.GradientTape();

        bool success = TreeToTensorConverter.TryEvaluate(
          tree, 
          inputFeatures, variables,
          updateVariableWeights, applyLinearScaling,
          eager,
          out Tensor prediction);
        if (!success)
          return (ISymbolicExpressionTree)tree.Clone();

        var loss = tf.reduce_mean(tf.square(target - prediction));

        progress?.Report(loss.ToArray<float>()[0]);

        var variablesList = variables.Values.SelectMany(x => x).ToList();
        var gradients = tape.gradient(loss, variablesList);
        
        optimizer.apply_gradients(zip(gradients, variablesList));
      }
      
      var cloner = new Cloner();
      var newTree = cloner.Clone(tree);
      var newConstants = variables.ToDictionary(
        kvp => (ISymbolicExpressionTreeNode)cloner.GetClone(kvp.Key),
        kvp => kvp.Value.Select(x => (double)(x.numpy().ToArray<float>()[0])).ToArray()
      );
      UpdateConstants(newTree, newConstants);


      return newTree;


      


//      //int numRows = rows.Count();






//      var variableLengths = problemData.AllowedInputVariables.ToDictionary(
//        var => var,
//        var => {
//          if (problemData.Dataset.VariableHasType<double>(var)) return 1;
//          if (problemData.Dataset.VariableHasType<DoubleVector>(var)) return problemData.Dataset.GetDoubleVectorValue(var, 0).Count;
//          throw new NotSupportedException($"Type of variable {var} is not supported.");
//        });

//      var variablesDict = problemData.AllowedInputVariables.ToDictionary(
//        var => var,
//        var => {
//          if (problemData.Dataset.VariableHasType<double>(var)) {
//            var data = problemData.Dataset.GetDoubleValues(var, rows).Select(x => (float)x).ToArray();
//            return tf.convert_to_tensor(np.array(data).reshape(new Shape(numRows, 1)), DataType);
//          } else if (problemData.Dataset.VariableHasType<DoubleVector>(var)) {
//            var data = problemData.Dataset.GetDoubleVectorValues(var, rows).SelectMany(x => x.Select(y => (float)y)).ToArray();
//            return tf.convert_to_tensor(np.array(data).reshape(new Shape(numRows, -1)), DataType);
//          } else  throw new NotSupportedException($"Type of the variable is not supported: {var}");
//        }
//      );

//      using var tape = tf.GradientTape(persistent: true);

//      bool success = TreeToTensorConverter.TryEvaluateEager(tree,
//        numRows, variablesDict,
//        updateVariableWeights, applyLinearScaling,
//        out Tensor prediction,
//        out Dictionary<Tensor, string> parameters, out List<ResourceVariable> variables);

//      //bool success = TreeToTensorConverter.TryConvert(tree,
//      //  numRows, variableLengths,
//      //  updateVariableWeights, applyLinearScaling,
//      //  out Tensor prediction,
//      //  out Dictionary<Tensor, string> parameters, out List<Tensor> variables);

//      if (!success)
//        return (ISymbolicExpressionTree)tree.Clone();

//      //var target = tf.placeholder(DataType, new Shape(numRows), name: problemData.TargetVariable);
//      var targetData = problemData.Dataset.GetDoubleValues(problemData.TargetVariable, rows).Select(x => (float)x).ToArray();
//      var target = tf.convert_to_tensor(np.array(targetData).reshape(new Shape(numRows)), DataType);
//      // MSE
//      var cost = tf.reduce_sum(tf.square(prediction - target));

//      tape.watch(cost);

//      //var optimizer = tf.train.AdamOptimizer((float)learningRate);
//      //var optimizer = tf.train.AdamOptimizer(tf.constant(learningRate, DataType));
//      //var optimizer = tf.train.GradientDescentOptimizer((float)learningRate);
//      //var optimizer = tf.train.GradientDescentOptimizer(tf.constant(learningRate, DataType));
//      //var optimizer = tf.train.GradientDescentOptimizer((float)learningRate);
//      //var optimizer = tf.train.AdamOptimizer((float)learningRate);
//      //var optimizationOperation = optimizer.minimize(cost);
//      var optimizer = keras.optimizers.Adam((float)learningRate);

//      #if EXPORT_GRAPH
//      //https://github.com/SciSharp/TensorFlow.NET/wiki/Debugging
//      tf.train.export_meta_graph(@"C:\temp\TFboard\graph.meta", as_text: false,
//        clear_devices: true, clear_extraneous_savers: false, strip_default_attrs: true);
//#endif

//      //// features as feed items
//      //var variablesFeed = new Hashtable();
//      //foreach (var kvp in parameters) {
//      //  var variable = kvp.Key;
//      //  var variableName = kvp.Value;
//      //  if (problemData.Dataset.VariableHasType<double>(variableName)) {
//      //    var data = problemData.Dataset.GetDoubleValues(variableName, rows).Select(x => (float)x).ToArray();
//      //    variablesFeed.Add(variable, np.array(data).reshape(new Shape(numRows, 1)));
//      //  } else if (problemData.Dataset.VariableHasType<DoubleVector>(variableName)) {
//      //    var data = problemData.Dataset.GetDoubleVectorValues(variableName, rows).SelectMany(x => x.Select(y => (float)y)).ToArray();
//      //    variablesFeed.Add(variable, np.array(data).reshape(new Shape(numRows, -1)));
//      //  } else
//      //    throw new NotSupportedException($"Type of the variable is not supported: {variableName}");
//      //}
//      //var targetData = problemData.Dataset.GetDoubleValues(problemData.TargetVariable, rows).Select(x => (float)x).ToArray();
//      //variablesFeed.Add(target, np.array(targetData));


//      List<NDArray> constants;
//      //using (var session = tf.Session()) {

//#if LOG_FILE
//        var directoryName = $"C:\\temp\\TFboard\\logdir\\manual_{DateTime.Now.ToString("yyyyMMddHHmmss")}_{maxIterations}_{learningRate.ToString(CultureInfo.InvariantCulture)}";
//        Directory.CreateDirectory(directoryName);
//        var costsWriter = new StreamWriter(File.Create(Path.Combine(directoryName, "Costs.csv")));
//        var weightsWriter = new StreamWriter(File.Create(Path.Combine(directoryName, "Weights.csv")));
//        var gradientsWriter = new StreamWriter(File.Create(Path.Combine(directoryName, "Gradients.csv")));
//#endif

//      //session.run(tf.global_variables_initializer());

//#if LOG_CONSOLE || LOG_FILE
//        var gradients = optimizer.compute_gradients(cost);
//#endif

//      //var vars = variables.Select(v => session.run(v, variablesFeed)[0].ToArray<float>()[0]).ToList();
//      //var gradient = optimizer.compute_gradients(cost)
//      //  .Where(g => g.Item1 != null)
//      //  //.Select(g => session.run(g.Item1, variablesFeed)[0].GetValue<float>(0)).
//      //  .Select(g => session.run(g.Item1, variablesFeed)[0].ToArray<float>()[0])
//      //  .ToList();

//      //var gradientPrediction = optimizer.compute_gradients(prediction)
//      //  .Where(g => g.Item1 != null)
//      //  .Select(g => session.run(g.Item1, variablesFeed)[0].ToArray<float>()[0])
//      //  .ToList();


//      //progress?.Report(session.run(cost, variablesFeed)[0].ToArray<float>()[0]);
//      progress?.Report(cost.ToArray<float>()[0]);


      


//#if LOG_CONSOLE
//        Trace.WriteLine("Costs:");
//        Trace.WriteLine($"MSE: {session.run(cost, variablesFeed)[0].ToString(true)}");

//        Trace.WriteLine("Weights:");
//        foreach (var v in variables) {
//          Trace.WriteLine($"{v.name}: {session.run(v).ToString(true)}");
//        }

//        Trace.WriteLine("Gradients:");
//        foreach (var t in gradients) {
//          Trace.WriteLine($"{t.Item2.name}: {session.run(t.Item1, variablesFeed)[0].ToString(true)}");
//        }
//#endif

//#if LOG_FILE
//        costsWriter.WriteLine("MSE");
//        costsWriter.WriteLine(session.run(cost, variablesFeed)[0].ToArray<float>()[0].ToString(CultureInfo.InvariantCulture));

//        weightsWriter.WriteLine(string.Join(";", variables.Select(v => v.name)));
//        weightsWriter.WriteLine(string.Join(";", variables.Select(v => session.run(v).ToArray<float>()[0].ToString(CultureInfo.InvariantCulture))));

//        gradientsWriter.WriteLine(string.Join(";", gradients.Select(t => t.Item2.Name)));
//        gradientsWriter.WriteLine(string.Join(";", gradients.Select(t => session.run(t.Item1, variablesFeed)[0].ToArray<float>()[0].ToString(CultureInfo.InvariantCulture))));
//#endif

//        for (int i = 0; i < maxIterations; i++) {
//          if (cancellationToken.IsCancellationRequested)
//            break;

          
//        var gradients = tape.gradient(cost, variables);
//        //optimizer.apply_gradients(gradients.Zip(variables, Tuple.Create<Tensor, IVariableV1>).ToArray());
//        optimizer.apply_gradients(zip(gradients, variables));
        

//        //session.run(optimizationOperation, variablesFeed);

//        progress?.Report(cost.ToArray<float>()[0]);
//        //progress?.Report(session.run(cost, variablesFeed)[0].ToArray<float>()[0]);

//#if LOG_CONSOLE
//          Trace.WriteLine("Costs:");
//          Trace.WriteLine($"MSE: {session.run(cost, variablesFeed)[0].ToString(true)}");

//          Trace.WriteLine("Weights:");
//          foreach (var v in variables) {
//            Trace.WriteLine($"{v.name}: {session.run(v).ToString(true)}");
//          }

//          Trace.WriteLine("Gradients:");
//          foreach (var t in gradients) {
//            Trace.WriteLine($"{t.Item2.name}: {session.run(t.Item1, variablesFeed)[0].ToString(true)}");
//          }
//#endif

//#if LOG_FILE
//          costsWriter.WriteLine(session.run(cost, variablesFeed)[0].ToArray<float>()[0].ToString(CultureInfo.InvariantCulture));
//          weightsWriter.WriteLine(string.Join(";", variables.Select(v => session.run(v).ToArray<float>()[0].ToString(CultureInfo.InvariantCulture))));
//          gradientsWriter.WriteLine(string.Join(";", gradients.Select(t => session.run(t.Item1, variablesFeed)[0].ToArray<float>()[0].ToString(CultureInfo.InvariantCulture))));
//#endif
//      }

//#if LOG_FILE
//        costsWriter.Close();
//        weightsWriter.Close();
//        gradientsWriter.Close();
//#endif
//      //constants = variables.Select(v => session.run(v)).ToList();
//      constants = variables.Select(v => v.numpy()).ToList();
//      //}

//      if (applyLinearScaling)
//        constants = constants.Skip(2).ToList();
//      var newTree = (ISymbolicExpressionTree)tree.Clone();
//      UpdateConstants(newTree, constants, updateVariableWeights);

//      return newTree;
    }

    private static void UpdateConstants(ISymbolicExpressionTree tree, Dictionary<ISymbolicExpressionTreeNode, double[]> constants) {
      foreach (var kvp in constants) {
        var node = kvp.Key;
        var value = kvp.Value;

        switch (node) {
          case ConstantTreeNode constantTreeNode:
            constantTreeNode.Value = value[0];
            break;
          case VariableTreeNodeBase variableTreeNodeBase:
            variableTreeNodeBase.Weight = value[0];
            break;
          case FactorVariableTreeNode factorVarTreeNode: {
            for (int i = 0; i < factorVarTreeNode.Weights.Length; i++) {
              factorVarTreeNode.Weights[i] = value[i];
            }
            break;
          }
        }
      }
    }

    //private static void UpdateConstants(ISymbolicExpressionTree tree, IList<NDArray> constants, bool updateVariableWeights) {
    //  int i = 0;
    //  foreach (var node in tree.Root.IterateNodesPrefix().OfType<SymbolicExpressionTreeTerminalNode>()) {
    //    if (node is ConstantTreeNode constantTreeNode) {
    //      constantTreeNode.Value = constants[i++].ToArray<float>()[0];
    //    } else if (node is VariableTreeNodeBase variableTreeNodeBase && updateVariableWeights) {
    //      variableTreeNodeBase.Weight = constants[i++].ToArray<float>()[0];
    //    } else if (node is FactorVariableTreeNode factorVarTreeNode && updateVariableWeights) {
    //      for (int j = 0; j < factorVarTreeNode.Weights.Length; j++)
    //        factorVarTreeNode.Weights[j] = constants[i++].ToArray<float>()[0];
    //    }
    //  }
    //}

    public static bool CanOptimizeConstants(ISymbolicExpressionTree tree) {
      return TreeToTensorConverter.IsCompatible(tree);
    }
  }
}